/*
AMD_15h_MSR_PSTATES
based on code of Copyright (c) Martin Kinkelin
*/

#pragma once

#include <vector>

#include "def.h"
using namespace Ndef;

#include "BaseService.h"
#include "Registry.h"
#include "TurboManager.h"
#include "../WinRing0/OlsApi.h"

#pragma comment(lib, "../WinRing0/WinRing0.lib")
//#pragma comment(lib, "../WinRing0/WinRing0x64.lib")



/*
 * To create your own service, customize the macros below, derive
 * from BaseService and set ServiceType to your service type.
 * Service.cpp should then include your header file instead of this
 * one.
 */

#define NAME "AMD_15h_MSR_PSTATES"
#define DISPLAY_NAME "AMD_15h_MSR_PSTATES service"
#define DESCRIPTION "Applies custom P-state settings."
#define START_TYPE SERVICE_AUTO_START // SERVICE_DEMAND_START, SERVICE_DISABLED
//#define ENABLE_DELAYED_AUTOSTART

#define REG_DIR "Software\\AMD_15h_MSR_PSTATES"

namespace Npstates{

class	AMD_15h_MSR_PSTATES_service;
typedef AMD_15h_MSR_PSTATES_service ServiceType;

class AMD_15h_MSR_PSTATES_service : public BaseService
{
public:
	AMD_15h_MSR_PSTATES_service(SERVICE_STATUS_HANDLE statusHandle, const SERVICE_STATUS& status) :
		BaseService(statusHandle, status),
		_key(NULL),
		_key_cur_set(NULL),
		is_check_CPU(false),
		is_mask_IDD(true),
		is_mask_K10_NbVID(true),
		is_avoid_wr_cur_pstate(false),
		is_sw_lowest_pstate(false)
	{
		SYSTEM_INFO info;
		GetSystemInfo(&info);

		_numCores = info.dwNumberOfProcessors;
	}

	virtual ~AMD_15h_MSR_PSTATES_service()
	{
		if (_key_cur_set != NULL){ RegCloseKey(_key_cur_set); _key_cur_set= NULL; }
		if (_key != NULL){ RegCloseKey(_key); _key= NULL; }
	}

public:

	bool OnInitialize()
	{
		//?is there no required winsys registry records (not installed)
		if (!ReadFromRegistry()) return false;

		//?check CPU
		//if (is_check_CPU){ if(!check_CPU()) return false; }

		//?WinRing0
		if (!InitializeOls()) return false;

		if (GetDllStatus() != 0)
		{
			DeinitializeOls();
			return false;
		}

		return true;
	}

	void OnRun()
	{
		// switch to a high process priority
		const HANDLE currentProcess = GetCurrentProcess();
		SetPriorityClass(currentProcess, HIGH_PRIORITY_CLASS);

		//Lock();{
		
		if (!StartWork()) Stop();
		
		//}Unlock();
	}

	void OnStop()
	{
		// stop the other threads, let them terminate and then delete them

		//off WinRing0
		DeinitializeOls();
	}

	void OnPowerEvent(DWORD eventType, PPOWERBROADCAST_SETTING setting)
	{
		if (eventType == PBT_APMRESUMEAUTOMATIC)
		{
			// the system just resumed from standby/hibernation
			// while the software (RAM) state is the same as before standby,
			// the hardware state isn't (CPU MSRs and fan controller registers)
			// => desync, start over
			//Lock();{
			
			// stop the other threads, let them terminate and then delete them

			//restart
			StartWork();

			//}Unlock();
		}
	}

private:
	HKEY		_key, _key_cur_set;

public:
	//CPU cores reported by Windows
	ureg32		_numCores;
	
	//in registry 
	//
	bool										is_check_CPU;
	//
	bool										is_mask_K10_NbVID;
	//
	bool										is_mask_IDD;
	//
	bool										is_avoid_wr_cur_pstate;
	//
	bool										is_sw_lowest_pstate;
	//subdir name for FSB_freq(0xxx)##user_suffix(????) containing current set of MSRS_low, MSRS_high
	enum { cur_set_name_strlen=1024 };
	char										cur_set_name[2*(cur_set_name_strlen+1)];
	//MSR records 
	std::vector< ureg32 >						MSRS_low;
	//<true> if MSRS_high exist for the MSRS_low
	std::vector< std::pair<bool,ureg32> >		MSRS_high;

private:
	/// <summary>Starts the work and returns true if there actually is work to do.</summary>
	bool StartWork()
	{
		return ApplyCustomPStates();
		//UnlockPStates();

		// the service needs to be kept alive to re-apply custom P-states when
		// the system resumes from standby
		//return true;
	}


	/// <summary>
	/// Saves custom P-state settings to the cores' MSRs.
	/// If a core is in a P-state being customized, the settings are applied immediately.
	/// </summary>
	bool ApplyCustomPStates()
	{
		//? user states exist
		if(!MSRS_low.size())return false;
		if(MSRS_low.size()!=MSRS_high.size())return false;

		//?try avoid writing to current selected P-state
		if(is_avoid_wr_cur_pstate){
		//
		const ureg32 NumBoostedStates = TurboManager::GetNumBoostedStates();

		//there are
		//MSRS_low.size() number of Pstates to change
		DWORD  low, high;

		//save P-state Current Limits
		RdmsrTx(0xC0010061u, &low, &high, 1);
		const ureg32  min_p_state_value= low & 0x07u;
		const ureg32  max_p_state_value= (low>>4) & 0x07u;
		
		//? at least 2 P-states enabled by power plan
		if( min_p_state_value == max_p_state_value )return false;

		// switch temporarily to a time-critical (absolutely highest) thread priority
		// (we try not to interfere with default C&Q here)
		HANDLE currentThread = GetCurrentThread();
		const int previousPriority = GetThreadPriority(currentThread);
		SetThreadPriority(currentThread, THREAD_PRIORITY_TIME_CRITICAL);

		//go to highest enabled P-state
		for( ureg32 core=0; core<_numCores; ++core ){
		const DWORD_PTR affinityMask = (DWORD_PTR)1 << core;

		// initiate switching to the temp state and immediately continue with the next core
		WrmsrTx(0xC0010062u, min_p_state_value, 0, affinityMask);
		}
		Sleep(3); // let transitions complete

		//setup all lower registers
		//from down to up, due to NBPs
		for( ureg32 core=0; core<_numCores; ++core ){
		//for( ureg32 idx=(NumBoostedStates+min_p_state_value+1); idx<MSRS_low.size(); ++idx){
		for( ureg32 idx=MSRS_low.size()-1; idx>=(NumBoostedStates+min_p_state_value+1); --idx){

		WriteMSR_C001_006x( 
			core, 
			idx, 
			MSRS_low[idx], 
			MSRS_high[idx],
			is_mask_K10_NbVID,
			is_mask_IDD
			);

		}}

		if(is_sw_lowest_pstate){
		//go to lowest enabled P-state
		for( ureg32 core=0; core<_numCores; ++core ){
		const DWORD_PTR affinityMask = (DWORD_PTR)1 << core;

		// initiate switching to the temp state and immediately continue with the next core
		WrmsrTx(0xC0010062u, max_p_state_value, 0, affinityMask);
		}
		Sleep(3); // let transitions complete

		//if(is_sw_lowest_pstate)
		}else{
		//go to (highest-1) enabled P-state
		for( ureg32 core=0; core<_numCores; ++core ){
		const DWORD_PTR affinityMask = (DWORD_PTR)1 << core;

		// initiate switching to the temp state and immediately continue with the next core
		WrmsrTx(0xC0010062u, min_p_state_value+1, 0, affinityMask);
		}

		Sleep(3); // let transitions complete
		}

		//setup all higher registers
		//from down to up, due to NBPs
		for( ureg32 core=0; core<_numCores; ++core ){
		//for( ureg32 idx=0; idx<=(NumBoostedStates+min_p_state_value); ++idx){
		for( ureg32 idxp1=(NumBoostedStates+min_p_state_value+1); idxp1; --idxp1){

		const ureg32 idx=idxp1-1;
		WriteMSR_C001_006x( 
			core, 
			idx, 
			MSRS_low[idx], 
			MSRS_high[idx],
			is_mask_K10_NbVID,
			is_mask_IDD
			);

		}}

		SetThreadPriority(currentThread, previousPriority);
		return true;

		//if(is_avoid_wr_cur_pstate)
		}else{

		// switch temporarily to a time-critical (absolutely highest) thread priority
		// (we try not to interfere with default C&Q here)
		HANDLE currentThread = GetCurrentThread();
		const int previousPriority = GetThreadPriority(currentThread);
		SetThreadPriority(currentThread, THREAD_PRIORITY_TIME_CRITICAL);

		//setup all registers
		//from down to up, due to NBPs
		for( ureg32 core=0; core<_numCores; ++core ){
		for( ureg32 idxp1=MSRS_low.size(); idxp1; --idxp1){

		const ureg32 idx=idxp1-1;
		WriteMSR_C001_006x( 
			core, 
			idx, 
			MSRS_low[idx], 
			MSRS_high[idx],
			is_mask_K10_NbVID,
			is_mask_IDD
			);

		}}

		SetThreadPriority(currentThread, previousPriority);
		return true;

		//else(is_avoid_wr_cur_pstate)
		}
	}

public:
	/// <summary>Reads the parameters from the registry.</summary>
	/// <returns>True if successful.</returns>
	bool ReadFromRegistry()
	{
		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_DIR, 0, KEY_QUERY_VALUE, &_key) != ERROR_SUCCESS)
			return false;

		//get is_check_CPU
		Registry::GetBool(_key, "is_check_CPU", is_check_CPU);

		//get is_mask_K10_NbVID
		Registry::GetBool(_key, "is_mask_K10_NbVID", is_mask_K10_NbVID);

		//get is_mask_IDD
		Registry::GetBool(_key, "is_mask_IDD", is_mask_IDD);

		//get is_avoid_wr_cur_pstate
		Registry::GetBool(_key, "is_avoid_wr_cur_pstate", is_avoid_wr_cur_pstate);

		//get 
		Registry::GetBool(_key, "is_sw_lowest_pstate", is_sw_lowest_pstate);

		//find cur set
		{ureg32 size=sizeof(cur_set_name);
		if( !Registry::GetCString(_key, "cur_set_name", cur_set_name, &size ) )return false;

		char reg_sub_dir[sizeof(REG_DIR)+sizeof(cur_set_name)+1];
		std::sprintf(reg_sub_dir,"%s\\%s",REG_DIR, cur_set_name);
		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, reg_sub_dir, 0, KEY_QUERY_VALUE, &_key_cur_set) != ERROR_SUCCESS)
			return false;
		}

		//get desired MSRS
		char	MSR_name[]="MSR_x00";
		for( ureg32 idx=0; idx<256; ++idx ){
		
		//
		{ureg32	low;
		std::sprintf(MSR_name,"MSR_L%02u",idx);
		if ( !Registry::GetDword(_key_cur_set, MSR_name, (DWORD&)low) )break;
		MSRS_low.push_back(low);
		}

		//
		{std::pair<bool,ureg32>	high(false,0);
		std::sprintf(MSR_name,"MSR_H%02u",idx);
		if ( Registry::GetDword(_key_cur_set, MSR_name, (DWORD&)(high.second)) )high.first=true;
		MSRS_high.push_back(high);
		}

		//for
		}

		RegCloseKey(_key_cur_set); _key_cur_set= NULL; 
		RegCloseKey(_key); _key= NULL; 
		return true;
	}

	/// <summary>ReadMSR</summary>
	//
	static void ReadMSR_C001_006x( 
			const ureg32 core, 
			const ureg32 index, 
			ureg32 &low, 
			std::pair<bool,ureg32> &high
		)
	{
		const unsigned int msrIndex = 0xC0010064u + index;
		const DWORD_PTR affinityMask = (DWORD_PTR)1 << core;

		DWORD tmp_low, tmp_high;
		RdmsrTx(msrIndex, &tmp_low, &tmp_high, affinityMask);

		low = tmp_low;
		high.second = tmp_high;
		high.first = true;
	}

	/// <summary>WriteMSR</summary>
	//
	static void WriteMSR_C001_006x( 
			const ureg32 core, 
			const ureg32 index, 
			const ureg32 low, 
			const std::pair<bool,ureg32> high,
			const bool is_mask_K10_NbVID,
			const bool is_mask_IDD
		)
	{
		const unsigned int msrIndex = 0xC0010064u + index;
		const DWORD_PTR affinityMask = (DWORD_PTR)1 << core;

		DWORD tmp_low, tmp_high;
		RdmsrTx(msrIndex, &tmp_low, &tmp_high, affinityMask);

		//const DWORD lowMsrMask = 0xFE40FFFFu;
		//lower = (lower & ~lowMsrMask) | (lowMsr & lowMsrMask);
		if(!is_mask_K10_NbVID)tmp_low = low;
		else{
		const DWORD lowMsrMask = 0xFE000000u;
		tmp_low = ( tmp_low & lowMsrMask) | ( low & (~lowMsrMask));
		}

		//
		if(high.first){
		if(!is_mask_IDD)tmp_high = high.second;
		else{
		const DWORD highMsrMask = 0x000003FFu;
		tmp_high = (tmp_high & highMsrMask) | (high.second & (~highMsrMask));
		}}

		//
		WrmsrTx(msrIndex, tmp_low, tmp_high, affinityMask);
	}
};

//namespace Npstates
}
